# nexus-server

## Introduction

**nexus-server** is the server component of the Nexus VDI (Virtual Desktop Infrastructure) project.

## Main concepts

**nexus-server** allows authentified clients to manage users, virtual machines (VMs) and templates.

What a user can or cannot do is determined by her/his capabilities. Additionally, each VM stores a list of capabilites defining who is allowed to do what.

Templates are base images upon which VMs are created. A VM is always created after a given template. To make an analogy with programming, a template is a class and a VM is an instance of that class.

### Filesystem layout

In **nexus-server**'s root directory, we find three directories:

- `bin` where binaries reside
- `config` where configuration files reside
- `data/templates` where VM templates reside
- `data/vms` where VMs reside
- `data/tmp` folder where temporary files are created (and deleted)

Here is the filesystem hierarchy for a system featuring one template and one VM:

```
├── bin
│   ├── genpwd
│   ├── nexus-server
│   ├── template_creator
│   └── vm_run
├── certs
│   ├── ca-cert.pem
│   ├── server-cert.pem
│   └── server-key.pem
├── config
│   └── users.json
└── data
    ├── templates
    │   └── 8ae56a30-3195-4aea-960d-abb45c47f99e
    │       ├── disk.qcow
    │       └── template.json
    ├── tmp
    └── vms
        └── 39d
            └── 431
                └── 39d431f2-bf92-4e82-b26d-d597bc1c8142
                    ├── disk.qcow
                    └── vm.json
```

## Configuration

The file `config/users.json` specifies the users of the system. Each user in `users.json` is defined by the following attributes:

- A unique email address (which acts as a unique ID)
- A first name
- A last name
- A salted hashed password (bcrypt algorithm)
- Capabilities, defining what the user is allowed to do

Here is an example of `users.json` defining user Jane Doe:

```{.json}
[
{
    "email": "janedoe@nexus.org",
    "firstname": "Jane",
    "lastname": "Doe",
    "pwd": "$2a$10$3MfOrdLLAEJFu0rZkGxGW.bHGiMC/uRD3B5igg5bvdwUedToRqOdO",
    "caps": {
        "TPL_CREATE": 1,
        "TPL_EDIT": 1,
        "TPL_DESTROY": 1,
        "TPL_LIST": 1,
        "USER_CREATE": 1,
        "USER_DESTROY": 1,
        "USER_LIST": 1,
        "USER_SET_CAPS": 1,
        "VM_CREATE": 1,
        "VM_SET_ACCESS": 1
    }
}
]
```

## Certificates

Certificates reside in the `certs` directory which contains the following files:

- `ca-cert.pem`: Certificate Authority public key
- `server-cert.pem`: Certificate Authority private key
- `server-key.pem`: server private key

These are required for TLS encryption (required by the rest API and remote desktop communication).

## Templates

Templates are base disk images used to create new VMs.

Templates are typically disk images of various OS users might want to create VMs from, e.g.:

- Ubuntu 20.04 server
- Ubuntu 20.04
- Xubuntu 22.04
- Debian 11
- Alpine
- etc.

Templates can be `public` or `private`:

- `public` templates are visible to everyone
- `private` templates are only visible to their owner (the user who created them)

VMs disks are based on templates. Each VM uses its own disk which is an overlay pointing to a template (base) disk image.
Hence, each VM only stores the differences from the template image. This scheme allows to save a tremendous amount of disk space.

### Templates layout

Each template is uniquely identified by a UUID (version 4) and resides in its own unique directory named after the template's UUID.

A template directory contains the following files:

- `template.json`: the template's configuration
- `disk.qcow`: the template's disk image in QCOW2 format

The `template.json` file defines template'ID, its name, its owner, its access type (`public` or `private`), and its creation time:

```{.json}
{
    "id": "12e064f2-2092-428e-9685-f3e573c18802",
    "name": "Xubuntu 22.04 with gcc toolchain and vscodium editor",
    "owner": "janedoe@nexus.org",
    "access": "public",
    "creationTime": "2022-07-31T16:42:17+01:00"
}
```

The template above is stored as follow in the filesystem:
```
data
└── templates
    └── 12e064f2-2092-428e-9685-f3e573c18802
        ├── disk.qcow
        └── template.json
```

## VMs

A VM is made out of two files, a config file and a disk file:

- `vm.json` describes the VM's configuration (ID, name, RAM, CPUs, owner, template, access)
- `disk.qcow` is the disk image that defines the VM's "physical disk"

The ID of a VM is a unique UUID (version 4).
VMs are stored in `data/vms` and each VM resides in its own directory.

Each VM has its own unique folder named after its UUID. This folder contains the VM config (`vm.json`) and disk image (`disk.qcow`). Each VM's folder resides in a 3-directory-deep file hierarchy, where:

- 1st level directory is characters [0-2] of the VM UUID (4096 directories max)
- 2nd level directory is characters [3-5] of the VM UUID (4096 directories max)
- 3rd level directory is the full VM UUID

As an example, the VM with ID `39d431f2-bf92-4e82-b26d-d597bc1c8142` is stored as:

```{.shell}
data
└── vms
    └── 39d
        └── 431
            └── 39d431f2-bf92-4e82-b26d-d597bc1c8142
                ├── disk.qcow
                └── vm.json
```

Here is the content of `vm.json`:

```{.json}
{
    "id": "39d431f2-bf92-4e82-b26d-d597bc1c8142",
    "owner": "janedoe@nexus.org",
    "name": "Systems Programming Class 2022 (Xubuntu 22.04)",
    "cpus": 2,
    "ram": 4096,
    "nic": "none",
    "usbDevs": [
        "067b:2303",
        "1052:b89a"
    ],
    "templateID": "12e064f2-2092-428e-9685-f3e573c18802",
    "access": {
        "lukeskywalker@force.net" : {
            "VM_LIST":1,
            "VM_ATTACH":1,
            "VM_START":1,
            "VM_STOP":1,
            "VM_REBOOT":1
        },
        "student@nexus.org" : {
            "VM_LIST":1,
            "VM_ATTACH":1
        }
    }
}
```

When the VM is created, an external snapshot of its template is created in the VM directory under the name `disk.qcow`. Thus, the VM's disk only stores differences from the template the VM is based on.

### VM creation

- When a VM is created, an "overlay" (snapshot) image baked by the template image is created
  - it means that every VM disk represents the differences from the template (base image)
- In the following example, `template.qcow` is the template image upon which the VM's disk `disk.qcow` is based:
  ```
  qemu-img create -F qcow2 -b template.qcow -f qcow2 disk.qcow
  ```

Remarks:

- To list the baking file of a disk:
  ```
  qemu-img info --backing-chain disk.qcow
  ```
- To create an empty disk of 200GB:
  ```
  qemu-img create -f qcow2 disk.qcow 200G
  ```

### Remote display

When a VM is started, **nexus-server** create a new thread in which a QEMU process (`qemu-system-x86_64`) is executed with an embedded spice server which allows anyone with the proper information (port and password) to remotely access the VM's display.

The spice server is configured to encrypt all communication channels through TLS.

### QEMU Guest Agent

To each running VM is associated a UNIX-domain socket used by **nexus-server** to interact with the VM through the QEMU Guest Agent Protocol. This mechanism is typically used to gracefully shutdown a VM, query the system (users, hardware, etc.), create files, execute commands, etc. The list of commands currently implemented by the QEMU Guest Agent Protocol is [described here](https://qemu.readthedocs.io/en/latest/interop/qemu-ga-ref.html).

However, for QEMU Guest Agent Protocol commands to work, the guest OS must be running the `qemu-guest-agent` systemd service (`qemu-ga` process). On a Ubuntu/Debian system, it can be installed with:
```
sudo apt-get install qemu-guest-agent
```

Consequently, it is highly recommended to install it in all templates as it gives you more features and control over your VMs.

## Access control

Access control [is described here](access_control.md).

## REST API

The REST API [is described here](rest_api.md).

## Building and installing nexus-server

To make the installation of **nexus-server** as easy as a possible, a `Makefile` is provided which automates most of the steps required to build and install the service.

Note that `guestfish` (package of the same name) is required. It should normally be ran as a non-root user. However, on Ubuntu, it requires root access due to a side-effect of Ubuntu's wrong file permissions with `/boot/vmlinuz*` files ([more info here](https://libguestfs.org/guestfs-faq.1.html)). Consequently, the command below must be run in order to restore the correct file permissions:
```
sudo chmod 0644 /boot/vmlinuz*
```

**BEWARE**: whenever a new kernel is installed, its permission **must be changed again**, otherwise importing or exporting files from VMs won't work!

### Prerequisites

The host machine on which **nexus-server** will be running requires the following software:

- A Linux kernel with the KVM module loaded
- QEMU 6.0.0 or newer (`qemu-system-x86` Ubuntu package)
- `qemu-img` to manipulate VM disk images (`qemu-utils` Ubuntu package)
- `guestfish` to export/import files from VM disk images (`guestfish` Ubuntu package)
- `uuidgen` to generate UUIDs from bash (`uuid-runtime` Ubuntu package)

Note that Ubuntu 20.04 ships with QEMU 4.2.1 which is too old, but Ubuntu 22.04 ships with QEMU 6.2.
The latest stable version of QEMU can be downloaded from [https://www.qemu.org/download/](https://www.qemu.org/download/). To compile QEMU from source, read the [Compiling QEMU from source](##Compiling-QEMU-from-source) section at the end of this document.

To get, build, and install **nexus-server**, these additionnal software are required:

- git (`git` Ubuntu package)
- golang 1.18 or newer (`golang-go` Ubuntu 22.04 packages)
- GNU make (`make` Ubuntu package)
- certtool (`gnutls-bin` Ubuntu package)

To install all the above-mentioned packages on a Ubuntu 22.04 system:
```
sudo apt-get install -y qemu-system-x86 qemu-utils guestfish uuid-runtime git golang-go make gnutls-bin
```

To make sure the KVM module is loaded into the kernel, run:
```
lsmod|grep kvm
```
If you don't see something similar to the output below (if you have an AMD CPU, the output is similar with `amd` instead of `intel`), it means the KVM module is not loaded into the kernel:
```
kvm_intel             294912  0
kvm                   782336  1 kvm_intel
```

To compile **nexus-server**, a Go compiler version 1.18 or above is required. Snap and flatpak package repositories provide recent enough versions. Alternatively, a go compiler toolchain from the official Go website can be obtained at [https://go.dev/dl/](https://go.dev/dl/). You also need `make` in order to execute the installation script.

The **nexus-server**'s source code must be obtained from its git repository with the following command using ssh key-pair authentication (as a non-root user):
```{.shell}
git clone ssh://git@ssh.hesge.ch:10572/flg_projects/nexus_vdi/nexus-server.git nexus-server.git
```

If you didn't set up a public key in your gitlab account, then you can obtain the sources through https with:
```{.shell}
git clone https://gitedu.hesge.ch/flg_projects/nexus_vdi/nexus-server.git nexus-server.git
```

Once the code repository cloned, go into `nexus-server.git`:
```
cd nexus-server.git
```

### Installation script

As mentioned above, the script is simply a `Makefile` located at the root of the git repository.

### Configuring certificates

Certificates are required to ensure encrypted communication (TLS) with nexus clients. The installation script creates the required certificates, but you must still edit a few fields specific to your system.

Edit `config/certs/nexus-server.info` and change the following fields to your liking: `organization`, `cn`, `dns_name` and `ip_address`.
Also, edit `config/certs/ca.info` and change the `cn` field accordingly.

### Configuring the service

By default, the service listens to port 1077. If you would like the service (API) to listen to a different port, edit `config/systemd/nexus-server.service` and  change the line ending in `-p 1077` with a port in the range [1025,1099] inclusive. If you wish to change this range, modify `APIPortMin` and `APIPortMax` in `src/server/consts/consts.go`. 

Furthermore, each running VM listens to a port randomly chosen in the range [1100,65535] inclusive. If you wish to change this range, modify `VMSpiceMinPort` and `VMSpiceMaxPort` in `src/server/consts/consts.go`.

### Building nexus-server and installing nexus-server

A `PREFIX` environment variable is used to specify where nexus-server must be installed. Let's assume we want `nexus-server` to reside in `/usr/local/`.

Run the following commands:
```
make build_srv
sudo -E make install_prod_srv PREFIX=/usr/local
```

The last command above:

- Creates a `nexus` user and group to run the service.
- Creates the file/directory hierarchy in the specified directory.
- Creates and signs the certificates.
- Add a crond task required by the service.
- Integrates the service with systemd.
- Creates a `users.json` file with user `admin@nexus.org`, featuring all capabilities. Its default password is `12345678`, therefore it is **highly** recommended to change it as described below.

The `genpwd` tools located in `PREFIX/nexus-server/bin/` can be used to obtain the hash of a password typed on the keyboard. This is very useful when wanting to update a user's password, such as `admin@nexus.org`  right after nexus' initial installation.

### Creating initial templates

To create VMs, at least one template is required. However, none is provided in the base install.

The `template_creator` script, located in `PREFIX/nexus-server/bin/`, can be used to create new templates from ISO images:
```
template_creator - a tool to create a VM template from an ISO image.
After the OS installation is completed, a tar archive is created in
the current directory. It can then be unarchived in the nexus-server's
templates directory.

Usage: template_creator name owner vis iso
name    name of the template
owner   email address of the template's owner
vis     visibility of the template, either "public" or "private"
iso     ISO image of the OS to install
```

For instance, this would create a template named "Xubuntu 22.04", owned by janedoe@nexus.org, visible to all users (i.e. `public`) and based on the ISO image located in `/tmp/xubuntu-22.04-desktop-amd64.iso`:
```
./template_creator "Xubuntu 22.04" janedoe@nexus.org public /tmp/xubuntu-22.04-desktop-amd64.iso
```

Once the command above is ran, a VM is launched in which you can proceed with the installation of the OS from the ISO image. Install the OS as you would on a physical PC. Once installed, you can beboot (the GUI lets you reboot the VM) and configure the OS to your liking. Once you're happy with the OS, close the VM window and `template_creator` will finish the creation of the template, by producing a `.tar` archive of the freshly created template.

You can now unarchive the archive in `PREFIX/nexus-server/data/templates`, set the proper owner, and delete the archive with (we assume the template archive is `7ac36ad7-a960-4c87-916a-4c1068d19b59.tar`):
```
sudo -s
cd PREFIX/nexus-server/data/templates
tar vfx PREFIX/nexus-server/bin/template_creator/7ac36ad7-a960-4c87-916a-4c1068d19b59.tar
chown -R nexus:nexus 7ac36ad7-a960-4c87-916a-4c1068d19b59
rm PREFIX/nexus-server/data/templates/7ac36ad7-a960-4c87-916a-4c1068d19b59.tar
```

Feel free to repeat the operation above with as many templates as you like.

### Updating a template

The `vm_run` script, located in `PREFIX/nexus-server/bin/`, is a convenience script that runs a VM on the specified image file(s) and it can be used to update a template's disk image. Here is its usage:
```
vm_run - launch a VM on the specified disk image(s).
The VM is configured with 4 CPUs and 4G of RAM.

Usage: vm_run -d disk -i iso
disk    the disk image to use (typically a .qcow file)
iso     the CD/DVD-ROM image to use (typically an .iso file)

At least one image must be specified.
```

Before updating a template's disk image, **make absolutely sure that it IS NOT used by any VMs**, otherwise all VMs based on this template will be corrupted!

### Copy a template from another server

To copy a template from another nexus server, `rsync` must be used instead of `scp` as it supports sparse files. Sparse files are files where consecutive zeros are not actually stored in data blocks. Template files (qcow) might appear much bigger than they really are as `ls` doesn't show the real block usage. To display the real size of a file, use `du -h FILE` instead.

Use the following recipe to copy a template using `rsync` over ssh (requires ssh key pairs):
```
# On the destination server, create the template directory
mkdir /path_to_templates/4913a2bb-edfe-4dfe-af53-38197a44523b

# On the source server, copy the template directory to the destination server
rsync -P --sparse -r 4913a2bb-edfe-4dfe-af53-38197a44523b  destination_server:/path_to_templates/4913a2bb-edfe-4dfe-af53-38197a44523b
 ```

### Controlling nexus-server with systemd

The nexus-server service can now be started with:
```
sudo systemctl start nexus-server
```

To check whether the service is running properly:
```
systemctl status nexus-server
```

You can then use any of the commands below to control your newly created `nexus-server` service. Only the `status` command does not require root privilege.

| Command                             | Description                                              |
|---                                  |---                                                       |
| `systemctl status nexus-server`     | display the service's status                             |
| `systemctl start nexus-server`      | start the service                                        |
| `systemctl stop nexus-server`       | stop the service                                         |
| `systemctl restart nexus-server`    | restart the service                                      |
| `systemctl reload nexus-server`     | reload the service                                       |
| `systemctl enable nexus-server`     | enable the service at boot time                          |
| `systemctl disable nexus-server`    | disable the service at boot time                         |
| `systemctl daemon-reload`           | reload the daemon (to call when service file is updated) |
| `journalctl -f --unit nexus-server` | display the service's log in "following" mode            |

### Public key for client access

Make sure to transmit the public key (certificate) located in `PREFIX/nexus-server/certs/ca-cert.pem` to any nexus client as they will need this key to authentify themselves and communicate with the server.

### Updating nexus-server

Whenever a new version of **nexus-server** is released, the binaries must be updated.

Go to the root source directory `nexus-server.git` and as a non-root user, build the binaries:
```
make build_srv
```

Then, as root (or sudo), stop the **nexus-server** systemd service:
```
sudo systemctl stop nexus-server
```

Next, as root (or sudo), update **nexus-server**:
```
sudo -E make update_srv PREFIX=/usr/local
```

Finally, as root (or sudo), start the **nexus-server** systemd service:
```
sudo systemctl start nexus-server
```

### Uninstalling nexus-server

**Beware this will delete ALL installed VMs and templates!**

First, as root (or sudo), stop the **nexus-server** systemd service:
```
sudo systemctl stop nexus-server
```

To uninstall **nexus-server**, go to its root source directory `nexus-server.git` and as sudo (or root), run:
```
sudo -E make uninstall_prod_srv PREFIX=/usr/local
```

## Resources

- QEMU Guest Agent Protocol
  - [QEMU Guest Agent Protocol Reference](https://qemu.readthedocs.io/en/latest/interop/qemu-ga-ref.html)

- QEMU Machine Protocol (QMP)
  - [Documentation/QMP](https://wiki.qemu.org/Documentation/QMP)
  - [QEMU QMP Reference Manual](https://qemu.readthedocs.io/en/latest/interop/qemu-qmp-ref.html)
  - [Features/CPUHotplug](https://wiki.qemu.org/Features/CPUHotplug)
  - [qmp-shell](https://github.com/0xef53/qmp-shell)

## USB redirection

### Introduction

The spice protocol implements USB redirection which allows USB devices plugged in the PC running the spice client (remote-viewer) to be accessible in the guest OS of the VM running on the server (the server runs QEMU with an embedded Spice server).

First, the `qemu-xhci` device must be added to the VM with:
```
-device qemu-xhci
```

Then, each USB device allowed to be accessed in the VM must be addedd to a filter list that forbids any other USB device. Each USB device is defined by two numbers, a vendor ID (`VID`) and a product ID (`PID`). Furthermore, one USB redirection, with a dedicated name, must be defined for each possible USB port.

Here is the general syntax for one redirection (`REDIR`) and one USB device (`VID`,`PID`):

```
-chardev spicevmc,name=usbredir,id=REDIR -device usb-redir,filter='-1:VID:PID:-1:1|-1:-1:-1:-1:0',chardev=REDIR
```

As many USB devices can be specified as necessary, but each redirection name must be unique.

In the syntax above, each USB device is represented by a filter which can match an arbitrary number of devices. The filter's syntax is as follow:
```
<class>,<vendor>,<product>,<version>,<allow>
```

Filters can be chained using the `|` operator. Here are a few examples:

- `-1:-1:-1:-1:1` → allows all USB devices
- `-1:-1:-1:-1:0` → forbids all USB devices
- `-1:0x123:0x456:-1:1|-1:-1:-1:-1:0` → allows USB device 0x123:0x456 and forbids everything else

Here is a real-world example that defines three USB redirections and each forbids everything except the three USB devices (VID:PID) below:

- A SanDisk Cruzer Blade USB key (0781:5567)
- A Prolific Technology USB/serial (067b:2303)
- An Atmel at91sam SAMBA bootloader (03eb:6124)

```
-device qemu-xhci -chardev spicevmc,name=usbredir,id=usbredir1 -device usb-redir,filter='-1:0x0781:0x5567:-1:1|-1:0x067b:0x2303:-1:1|-1:0x03eb:0x6124:-1:1|-1:-1:-1:-1:0',chardev=usbredir1 -chardev spicevmc,name=usbredir,id=usbredir2 -device usb-redir,filter='-1:0x0781:0x5567:-1:1|-1:0x067b:0x2303:-1:1|-1:0x03eb:0x6124:-1:1|-1:-1:-1:-1:0',chardev=usbredir2 -chardev spicevmc,name=usbredir,id=usbredir3 -device usb-redir,filter='-1:0x0781:0x5567:-1:1|-1:0x067b:0x2303:-1:1|-1:0x03eb:0x6124:-1:1|-1:-1:-1:-1:0',chardev=usbredir3
```

## Compiling QEMU from source

1. Retrieve QEMU' source code from [https://www.qemu.org/download/](https://www.qemu.org/download/)
1. Decompress the archive and go into QEMU' source directory:
   ```
   tar vfxJ qemu-7.0.0.tar.xz
   cd qemu-7.0.0
   ```
1. Install the required dependencies:
   ```
   sudo apt-get install -y build-essential meson libspice-server-dev libattr1-dev libcap-ng-dev libusb-1.0-0-dev libusbredirparser-dev
   ```
1. Enable spice (for remote desktop), virtfs (for shared folders), libusb (for usb passthrough) and usb-redir (for usb redirection over the network) in QEMU's config:
   ```
   ./configure --enable-spice --enable-virtfs --enable-libusb --enable-usb-redir
   ```
1. Compile QEMU:
   ```
   make
   ```
1. Install QEMU system-wide into `/usr/local/bin`:
   ```
   make install
   ```

<!--
### HOWTO: Exporting a VM disk image

Given each VM's disk is an overlay baked by a template's disk image, exporting a VM's disk image into a dedicated and independant file is slightly tricky. Below are the steps to do so.

Let:

- `base.qcow` be the template's disk image
- `overlay.qcow` be the VM's disk image (baked by `base.qcow`)

We therefore have the following relationship:

```
base.qcow <-- overlay.qcow
```

Steps:

1) Copy the overlay into a new one:
   ```
   cp overlay.qcow new_overlay.qcow
   ```
2) Rebase the new overlay in order to become a standalone disk file:
   ```
   qemu-img rebase -b "" new_overlay.qcow
   ```
3) Move the new overlay disk into the template directory:
   ```
   mv new_overlay.qcow $NEXUS_DATA_PATH/templates/my_new_template/
   ```

After completion of these 3 steps, `new_overlay.qcow` is equivalent to `base.qcow` + `overlay.qcow`.

Note that step 2) is very slow and can take several minutes given the template file can be very large (> 10GB).
-->
