package exec

import (
    "errors"
    "os/exec"
    "strings"
    "strconv"
    "runtime"
)

const (
    remoteViewerBinary = "remote-viewer"
)

func CheckRemoteViewer() error {

    os := runtime.GOOS
    output, err := exec.Command(remoteViewerBinary, "--version").Output()
    if os == "linux" {
        if err != nil {
            return errors.New(remoteViewerBinary+" is required but not found. On Ubuntu/Debian, "+remoteViewerBinary+" is part of the virt-viewer package that can be installed with \"sudo apt-get install virt-viewer\".")
        }
    } else if os == "darwin" {
        if err != nil {
            return errors.New(remoteViewerBinary+" is required but not found. On OSX, "+remoteViewerBinary+" is part of the virt-viewer package from MacPorts that can be installed from here: https://ports.macports.org/port/virt-viewer/")
        }
    } else {  // must be "windows"
        if err != nil {
            return errors.New(remoteViewerBinary+" is required but not found. Make sure to add it to the system PATH. On Windows, "+remoteViewerBinary+" is part of the virt-viewer MSI package that can be installed from here: https://virt-manager.org/download")
        }
    }
    out := string(output)
    lines := strings.Split(out, "\n")
    fields := strings.Split(lines[0], " ")
    if len(fields) < 3 {
        return errors.New("Failed extracting "+remoteViewerBinary+" version number!")
    }
    // This is what we should get if remote-viewer is working properly.
    if fields[0] == remoteViewerBinary && fields[1] == "version" {
        return nil
    }
    return errors.New("Failed extracting "+remoteViewerBinary+" version number!")
}

// Run remote-viewer.
// name is the VM's name.
// port is the port the VM spice server is listening to.
// pwd is the spice password to connect to the running VM.
func RunRemoteViewer(hostname string, cert string, name string, port int, pwd string, fullscreen bool) ([]byte, error) {
    portStr := strconv.Itoa(port)
    spice := "spice://"+hostname+"?tls-port="+portStr+"&password="+pwd
    spiceCert := "--spice-ca-file="+cert
    spiceSecure := "--spice-secure-channels=all"
    // To exit fullscreen mode, either:
    // - move the mouse to the middle of the top of the screen.
    // - use ctrl+F12 as specified below
    // To activate kiosk mode, add these options: "-k --kiosk-quit=on-disconnect"
    var c *exec.Cmd
    if fullscreen {
        c = exec.Command("remote-viewer", spice, spiceCert, spiceSecure, "-t", name, "--hotkeys=toggle-fullscreen=ctrl+f12", "-f")
    } else {
        c = exec.Command("remote-viewer", spice, spiceCert, spiceSecure, "-t", name, "--hotkeys=toggle-fullscreen=ctrl+f12")
    }
    return c.CombinedOutput()
}
