package main

import (
    "os"
    "path"
    "strings"
    "syscall"
    "os/signal"
    "nexus-common/utils"
    u "nexus-client/utils"
    g "nexus-client/globals"
    "nexus-client/defaults"
    "nexus-client/version"
    "nexus-client/cmd"
    "nexus-client/cmdMisc"
    "nexus-client/cmdLogin"
    "nexus-client/cmdToken"
    "nexus-client/cmdUser"
    "nexus-client/cmdVM"
    "nexus-client/cmdVersion"
    "nexus-client/cmdTemplate"
    "github.com/go-resty/resty/v2"
)

var cmdList = []cmd.Command {
    &cmdMisc.HelpHeader{"!═════╡ GENERAL commands ╞═════════════════════════════════════════════════════════════════"},
    // Commands in this block are specific to nexus-cli
    &cmdLogin.Login{"login"},
    // End of nexus-cli specific commands

    &cmdToken.Refresh{"refresh"},
    &cmdVersion.Version{"version"},

    &cmdMisc.HelpHeader{"!═════╡ USER commands ╞═════════════════════════════════════════════════════════════════"},
    &cmdUser.Whoami{"whoami"},
    &cmdUser.UpdatePwd{"passwd"},
    &cmdUser.List{"userlist"},
    &cmdUser.Add{"usercreate"},
    &cmdUser.Del{"userdel"},
    &cmdUser.SetCaps{"usersetcaps"},

    &cmdMisc.HelpHeader{"!═════╡ TEMPLATE commands ╞═════════════════════════════════════════════════════════════════"},
    &cmdTemplate.Create{"tplcreate"},
    &cmdTemplate.Del{"tpldel"},
    &cmdTemplate.Edit{"tpledit"},
    &cmdTemplate.ExportDisk{"tplexportdisk"},
    &cmdTemplate.List{"tpllist"},

    &cmdMisc.HelpHeader{"!═════╡ VM commands ╞═════════════════════════════════════════════════════════════════"},
    &cmdVM.AddAccess{"vmaddaccess"},
    &cmdVM.AttachAsync{"vmattach"},
    &cmdVM.Create{"vmcreate"},
    &cmdVM.Creds2pdf{"vmcreds2pdf"},
    &cmdVM.Creds2csv{"vmcreds2csv"},
    &cmdVM.Del{"vmdel"},
    &cmdVM.DelAccess{"vmdelaccess"},
    &cmdVM.Edit{"vmedit"},
    &cmdVM.ExportDir{"vmexportdir"},
    &cmdVM.ImportDir{"vmimportdir"},
    &cmdVM.Stop{"vmkill"},
    &cmdVM.List{"vmlist"},
    &cmdVM.Reboot{"vmreboot"},
    &cmdVM.Shutdown{"vmshutdown"},
    &cmdVM.Start{"vmstart"},
    &cmdVM.StartWithCreds{"vmstartwithcreds"},
}

func run() int {
    var appname = path.Base(os.Args[0])
    u.PrintlnErr(appname+" version "+version.Get().String())

    if len(os.Args) < 2 {
        u.PrintlnErr("USAGE: "+appname+" CMD")
        u.PrintlnErr("CMD is the Command to run.")
        u.PrintlnErr("All commands besides login read the access token from the env. variable \"NEXUS_TOKEN\".")
        u.PrintlnErr("To obtain an access token, use the \"login\" command.")
        cmd.Help(cmdList)
        return 1
    }

    serverEnvVar, found := os.LookupEnv(g.ENV_NEXUS_SERVER)
    if !found {
        serverEnvVar = defaults.NexusServer
        // u.PrintlnErr("Environment variable \""+g.ENV_NEXUS_SERVER+"\" must be set!")
        // u.PrintlnErr("It defines the nexus server to connect to along the port number.")
        // u.PrintlnErr("Example: export "+g.ENV_NEXUS_SERVER+"=192.168.1.42:1077")
        // return 1
    }

    var err error

    certEnvVar, found := os.LookupEnv(g.ENV_NEXUS_CERT)
    if !found {
        certEnvVar, err = defaults.CreateCert()
        if err != nil {
            u.PrintlnErr("Failed creating certificate from embedded certificate: ")
            return 1
        }
        defer os.Remove(certEnvVar)

        // This thread acts as a signal handler for SIGINT or SIGTERM.
        // When one of these signals is received, the temporary certificate file is deleted.
        // Without this "handler", the temporary certificate file wouldn't be deleted.
        go func(certFile string) {
            // Wait on dedicated signals.
            sigs := make(chan os.Signal, 1)
            signal.Notify(sigs, syscall.SIGINT, syscall.SIGTERM, syscall.SIGHUP)

            for {
                sig := <-sigs  // blocks on any of the above signals.
                u.Println("Caught signal ("+sig.String()+")")
                break;
            }

            os.Remove(certFile)
            os.Exit(1)
        }(certEnvVar)

        // u.PrintlnErr("Environment variable \""+g.ENV_NEXUS_CERT+"\" must be set!")
        // u.PrintlnErr("It specifies the path to the public certificate required for encrypted communications (TLS) with the nexus server.")
        // u.PrintlnErr("Example: export "+g.ENV_NEXUS_CERT+"=ca-cert.pem")
        // return 1
    }

    if !utils.FileExists(certEnvVar) {
        u.PrintlnErr("Failed reading certificate \""+certEnvVar+"\"!")
        return 1
    }

    parts := strings.Split(serverEnvVar, ":")
    hostname := parts[0]

    client := resty.New()
    client.SetRootCertificate(certEnvVar)
    host := "https://"+serverEnvVar

    g.Init(hostname, host, certEnvVar, client)

    // Checks the client version is compatible with the server's API.
    if !cmdVersion.CheckServerCompatibility(appname) {
        return 1
    }

    cmdName := os.Args[1]
    cmdArgs := os.Args[2:]

    found, cmd := cmd.Match(cmdName, cmdList)
    if found {
        if cmdName != "login" {
            token, found := os.LookupEnv(g.ENV_NEXUS_TOKEN)
            if !found || len(token) == 0 {
                u.PrintlnErr("Environment variable \""+g.ENV_NEXUS_TOKEN+"\" must be set!")
                return 1
            }
            client.SetHeader("Content-Type", "application/json")
            client.SetHeader("Authorization", "Bearer "+token)
        }
        return cmd.Run(cmdArgs)
    }

    u.PrintlnErr("Invalid Command")
    return 1
}

func main() {
    os.Exit(run())
}
