package utils

import (
    "os"
	"io"
	"io/ioutil"
    "fmt"
    "io/fs"
    "bytes"
    "errors"
    "strings"
    "net/mail"
    "archive/tar"
    "path/filepath"
    "compress/gzip"
    "github.com/google/uuid"
)

func Print(a ...any) {
    os.Stdout.WriteString(fmt.Sprint(a ...))
}

func Println(a ...any) {
    Print(a...)
    Print("\n")
}

func PrintErr(a ...any) {
    os.Stderr.WriteString(fmt.Sprint(a ...))
}

func PrintlnErr(a ...any) {
    PrintErr(a...)
    PrintErr("\n")
}

// Creates a tar.gz archive of dir and all its files and subdirectories.
// Note: dir can also be a file.
// Source code slightly modified from: https://gist.github.com/mimoo/25fc9716e0f1353791f5908f94d6e726
func TarGzDir(dir, archive string) error {
    _, err := os.Stat(dir)
    if errors.Is(err, fs.ErrNotExist) {
        return errors.New("Error: \""+dir+"\" does not exist");
    }

    var buf bytes.Buffer

    gzWriter := gzip.NewWriter(&buf)
    tarWriter := tar.NewWriter(gzWriter)

    // Walks through every file in the directory
    filepath.Walk(dir, func(file string, fi os.FileInfo, err error) error {
        header, err := tar.FileInfoHeader(fi, file)
        if err != nil {
            return err
        }

        fileRel := strings.TrimPrefix(file, dir)
        fileRel = filepath.Join(filepath.Base(dir), fileRel)
        header.Name = filepath.ToSlash(fileRel)

        if err := tarWriter.WriteHeader(header); err != nil {
            return err
        }
        // if not a dir, write file content
        if !fi.IsDir() {
            data, err := os.Open(file)
            if err != nil {
                return err
            }
            if _, err := io.Copy(tarWriter, data); err != nil {
                return err
            }
        }
        return nil
    })

    if err := tarWriter.Close(); err != nil {
        return err
    }
    if err := gzWriter.Close(); err != nil {
        return err
    }

    fileToWrite, err := os.OpenFile(archive, os.O_CREATE|os.O_RDWR, os.FileMode(0750))
    if err != nil {
        return err
    }
    if _, err := io.Copy(fileToWrite, &buf); err != nil {
        return err
    }

    return nil
}

func GetRandomTempFilename() (string, error) {
    tempDir := os.TempDir()
    uuid, err := uuid.NewRandom()
    if err != nil {
        return "", errors.New("Failed creating random UUID: "+err.Error())
    }
    randName := "temp_"+uuid.String()
    return filepath.Join(tempDir, randName), nil
}

func IsEmail(email string) bool {
    _, err := mail.ParseAddress(email)
    return err == nil
}

// Convert a string of USB devices of the form "1fc9:001d,067b:2303"
// into a slice of string where each element is a string of the form "1fc9:001d".
// Returns an empty slice if the input string is "none".
func Str2UsbDevices(s string) []string {
    usbDevs := []string{}
    if s != "none" {
        devs := strings.Split(s, ",")  // Extracts USB devices
        for _, dev := range devs {
            usbDevs = append(usbDevs, dev)
        }
    }
    return usbDevs
}

// Returns the content of file as a string
func FileToString(file string) (string, error) {
	content, err := ioutil.ReadFile(file)
	if err != nil {
		return "", err
	}
	return string(content), nil
}

// Removes an element from a string array at a given index
func RemoveArgAtIndex(slice []string, index int) []string {
    return append(slice[:index], slice[index+1:]...)
}

// TODO: better way of appending a portable "new line" to a string
func AppendNewLine(s string) string {
    var newLine string
    newLine = fmt.Sprintln(newLine, "")
    return s+newLine
}
