package exec

import (
    "os/exec"
    "strings"
    "errors"
    "strconv"
    "nexus-server/utils"
    "nexus-server/config"
)

const (
    qemusystemMinVersion = 6
)

var conf = config.GetInstance()

// Check qemu-system-x86_64 is available and the version is recent enough.
func CheckQemuSystem() error {
    output, err := exec.Command(conf.Tools.QemuSystem, "-version").Output()
    if err != nil {
        return errors.New(conf.Tools.QemuSystem+" is required but not found. On Ubuntu/Debian, it can be installed with \"sudo apt-get install qemu-system-x86\".")
    }
    out := string(output)
    lines := strings.Split(out, "\n")
    fields := strings.Split(lines[0], " ")
    if len(fields) < 4 {
        return errors.New("Failed extracting "+conf.Tools.QemuSystem+" version number!")
    }
    v := fields[3]
    majorVersionNumber, err := strconv.Atoi(string(v[0]))
    if err != nil {
        return errors.New("Failed extracting "+conf.Tools.QemuSystem+" version number: "+err.Error())
    }
    if majorVersionNumber < qemusystemMinVersion {
        return errors.New(conf.Tools.QemuSystem+" "+strconv.Itoa(qemusystemMinVersion)+".0.0 or newer is required. Found version "+v+".")
    }
    return nil
}

// Creates a qemu-system command.
// The nic argument must be either "none" or "user".
func NewQemuSystem(qgaSock string, cpus, ram int, nic string, usbDevs []string, diskFile string, spicePort int, secretPwdFile, certDir string) (*exec.Cmd, error) {
    // Network config
    nicArgs := "none"
    if nic == "user" {
        addr, err := utils.RandMacAddress()
        if err != nil {
            return nil, errors.New("Failed executing VM: MAC address generation error")
        }
        nicArgs = "user,mac="+addr+",model=virtio-net-pci"
    }

    // If there are USB devices, generates the QEMU command line to support them
    usb := []string{}

    if len(usbDevs) > 0 {
        // A qemu-xhci device is required
        usb = append(usb, "-device", "qemu-xhci")

        // Generate the USB devices' filter list, for instance:
        // "-1:0x0781:0x5567:-1:1|-1:0x067b:0x2303:-1:1|-1:-1:-1:-1:0"
        var filter strings.Builder
        for _, dev := range usbDevs {
            filter.WriteString("-1:")
            ids := strings.Split(dev, ":")  // Extracts vendorID (vid) and productID (pid)
            usbDev := "0x"+ids[0]+":0x"+ids[1]
            filter.WriteString(usbDev)
            filter.WriteString(":-1:1|")
        }
        filter.WriteString("-1:-1:-1:-1:0")

        // Generate QEMU's arguments for the USB devices
        for i, _ := range usbDevs {
            idx := strconv.Itoa(i+1)
            usb = append(usb, "-chardev", "spicevmc,name=usbredir,id=usbredir"+idx)
            usb = append(usb, "-device", "usb-redir,filter="+filter.String()+",chardev=usbredir"+idx)
        }
    }

    // Enable KVM
    args := []string{"-enable-kvm"}
    // Suppresses several non-needed backends: serial, parallel, floppy, CD-ROM, SD card, etc.
    // More info here: https://gitlab.com/qemu-project/qemu/-/issues/322
    args = append(args, "-nodefaults")
    // CPU
    args = append(args, "-cpu", "host", "-smp", "cpus="+strconv.Itoa(cpus))
    // RAM
    args = append(args, "-m", strconv.Itoa(ram))
    // Harddrive
    args = append(args, "-drive", "file="+diskFile+",index=0,media=disk,format=qcow2,discard=unmap,detect-zeroes=unmap,if=virtio")
    // Display: use virtio instead of qxl. We've experienced recurrent X11 session crashes with qxl
    // while virtio seems much more robust. On the other hand, it seems that virtio uses slightly
    // more bandwidth.
    args = append(args, "-vga", "virtio")
    // Virtio serial
    args = append(args, "-device", "virtio-serial-pci")
    // Secret password for SPICE
    args = append(args, "-object", "secret,id=sec,file="+secretPwdFile+",format=base64")
    // SPICE
    args = append(args, "-spice", "tls-port="+strconv.Itoa(spicePort)+",password-secret=sec,x509-dir="+certDir+",image-compression=glz,jpeg-wan-compression=always,zlib-glz-wan-compression=always")
    args = append(args, "-chardev", "spicevmc,id=spicechannel0,name=vdagent")

    // Network bandwidth optimization
    // I measured network traffic (with "btop") while running the "cage" xscreensaver with these options:
    // 1) -vga virtio => 5.8-7.3 MiB
    // 2) -vga qxl => 5.2-6.1 MiB
    // 3) -vga -spice virtio image-compression=glz,jpeg-wan-compression=always,zlib-glz-wan-compression=always => 4-5 MiB
    // 4) -vga qxl -spice image-compression=glz,jpeg-wan-compression=always,zlib-glz-wan-compression=always => 3.5-4 MiB
    // In summary, the lowest bandwidth results are obtained with options 4)
    // Finally, that I didn't notice any visual quality difference between these 4 options.

    // SPICE: allows copy/paste between guest & client
    args = append(args, "-device", "virtserialport,chardev=spicechannel0,name=com.redhat.spice.0")
    // Network
    args = append(args, "-nic", nicArgs)
    // QEMU Guest Agent
    args = append(args, "-device", "virtio-serial", "-device", "virtserialport,chardev=qga0,name=org.qemu.guest_agent.0", "-chardev", "socket,path="+qgaSock+",server=on,wait=off,id=qga0")
    // USB redirection
    args = append(args, usb...)
    // Sound support
    args = append(args, "-device", "intel-hda", "-device", "hda-duplex")

    // To share a folder with the host:
    // - Add QEMU args: virtfs local,path=/tmp/pipo,mount_tag=sharedfs,security_model=none
    // - In the guest: mkdir shared && sudo mount -t 9p sharedfs shared

    cmd := exec.Command(conf.Tools.QemuSystem, args...)
    return cmd, nil
}
