package router

import (
    "time"
    "errors"
    "net/http"
    "nexus-server/users"
    "nexus-common/params"
    "golang.org/x/crypto/bcrypt"
    "github.com/golang-jwt/jwt"
    "github.com/labstack/echo/v4"
    "github.com/labstack/echo/v4/middleware"
    "github.com/go-playground/validator/v10"
    "github.com/sethvargo/go-password/password"
)

// jwt.StandardClaims is added as an embedded type to provide fields like user email and expiration time.
type CustomClaims struct {
    Email string
    jwt.StandardClaims
}

type Auth struct {
    users *users.Users
    tokenAccess middleware.JWTConfig
}

var loginAttempts map[string]int  // Keeps track of the number of wrong logins.
var loginLocked map[string]time.Time   // Keeps track of when an account was locked.
var jwtSecretKey string

// Creates an Auth object.
func NewAuth(u *users.Users) (*Auth, error) {
    loginAttempts = make(map[string]int)
    loginLocked = make(map[string]time.Time)

    // TODO: better to store jwtSecretKey in an env. variable?
    // Generates a 64 characters long key with 20 digits, 0 symbols,
    // allowing upper and lower case letters, and allowing repeat characters.
    secret, err := password.Generate(64, 20, 0, false, true)
    if err != nil {
        return nil, errors.New("Secret generation error: "+err.Error())
    }
    jwtSecretKey = secret

    return &Auth {
        users : u,
        tokenAccess : middleware.JWTConfig {
            ContextKey: "user",
            SigningKey: []byte(jwtSecretKey),
            Claims:     &CustomClaims{},
            ErrorHandlerWithContext: func(err error, c echo.Context) error {
                return echo.NewHTTPError(http.StatusUnauthorized, "access denied")
            },
        },
    }, nil
}

// Token used by the JWT middleware to protect access to the specified group of routes.
func (auth *Auth)GetTokenAccess() middleware.JWTConfig {
    return auth.tokenAccess
}

// curl --cacert ca.pem -X POST https://localhost:1077/login -H 'Content-Type: application/json' -d '{"email": "johndoe@nexus.org", "pwd":"the_user_password"}'
// Returns a jwt token if authentication was successful:
// {"token":"<jwt token>"}
func (auth *Auth)Login(c echo.Context) error {
    // Deserializes the JSON body and checks its validity.
    reqUser := new(params.Login)
    if err := decodeJson(c, &reqUser); err != nil {
        return echo.NewHTTPError(http.StatusInternalServerError, err.Error())
    }
    if err := validator.New(validator.WithRequiredStructEnabled()).Struct(reqUser); err != nil {
        return echo.NewHTTPError(http.StatusBadRequest, err.Error())
    }

    // User exists?
    user, err := auth.users.GetUserByEmail(reqUser.Email)
    if err != nil {
        return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
    }

    // Is user locked?
    lockedTime, exists := loginLocked[user.Email]
    if exists {
        // If locked time has elapsed, unlock the account.
        t := time.Now()
        elapsedTime := t.Sub(lockedTime)
        if elapsedTime.Hours() > conf.Auth.LockedTimeInHours {
            delete(loginLocked, user.Email)
            delete(loginAttempts, user.Email)
        } else {
            log.Info("Account ", user.Email ," is still locked")
            return echo.NewHTTPError(http.StatusUnauthorized, "Account is locked")
        }
    }

    // Checks the user password is correct.
    // If a user enters the wrong password too many times, the account will be locked for a number of hours.
    if err := bcrypt.CompareHashAndPassword([]byte(user.Pwd), []byte(reqUser.Pwd)); err != nil {
        // Stores the number of attempted logins.
        count, exists := loginAttempts[user.Email]
        if exists {
            loginAttempts[user.Email] = count+1
            if count >= conf.Auth.MaxLoginAttempts {
                // Locks the account.
                loginLocked[user.Email] = time.Now()
                log.Info("Account ", user.Email ," is locked: too many wrong password attempts!")
                return echo.NewHTTPError(http.StatusUnauthorized, "Account is locked")
            }
        } else {
            loginAttempts[user.Email] = 1
        }

        // If the two passwords don't match, return a 401 status.
        return echo.NewHTTPError(http.StatusUnauthorized, "Invalid password")
    } else {
        loginAttempts[user.Email] = 0
    }

    // Retrieves user from its email.
    user, err = auth.users.GetUserByEmail(reqUser.Email)
    if err != nil {
        return echo.NewHTTPError(http.StatusUnauthorized, "User not found")
    }

    token, err := createToken(user.Email)
    if err != nil {
        return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
    }

    return c.JSON(http.StatusOK, echo.Map{"token": token})
}

// Creates a JWT token and claims with the expiration time for the given user.
func createToken(userEmail string) (string, error) {
    expirationTime := time.Now().Add(time.Duration(conf.Auth.AccessTokenExpirationTimeInHours)*time.Hour)

    // Creates the claims
    claims := &CustomClaims{
        Email: userEmail,
        StandardClaims: jwt.StandardClaims {
            // JWT expresses expiration time in Unix milliseconds
            ExpiresAt: expirationTime.Unix(),
        },
    }

    // Declares the token with the algorithm used for signing, and the claims.
    token := jwt.NewWithClaims(jwt.SigningMethodHS256, claims)

    // Signs the token and gets a string version of the token.
    secret := []byte(jwtSecretKey)
    tokenString, err := token.SignedString(secret)
    if err != nil {
        return "", err
    }

    return tokenString, nil
}

// Returns user email from the Echo context.
// If an error occured, returns the error.
func GetUserFromContext(c echo.Context) (string, error) {
    if c.Get("user") == nil {
        return "", errors.New("user not authenticated")
    }

    user := c.Get("user").(*jwt.Token)
    claims := user.Claims.(*CustomClaims)

    return claims.Email, nil
}

/*
// Example of middleware
// Such a middleware could be used to handle a logout feature.
// We would keep a map of invalid tokens (the ones that logged out), and deny access to such tokens.
// Question: Entries in the map would have to be cleared at times to prevent it from growing too much.
//           This could be done when the token has been in the map for longer than its expiration time.

usersGroup.Use(auth.OnlyJoe)  // apply the middleware (to be added in router.go)

func OnlyJoe(next echo.HandlerFunc) echo.HandlerFunc {
    return func(c echo.Context) error {
        // Gets user token from context.
        if c.Get("user") == nil {
            return echo.NewHTTPError(http.StatusUnauthorized, "invalid user token\n")
        }
        user := c.Get("user").(*jwt.Token)
        claims := user.Claims.(*CustomClaims)
        username := claims.Username
        if username != "joe" {
            return echo.ErrUnauthorized
        }
        return next(c)
    }
}
*/
